##############################################################################
# Copyright (c) 2008, Mooneer Salem
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Kite Language organization nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##############################################################################

##############################################################################
# Tree node object (not intended for public use)
##############################################################################
class _TreeNode [
    property left,
    property right,
    property key,
    property item,
    property parent,

    construct (k, v, p) [
        this.key = k;
        this.item = v;
        this.left = null;
        this.right = null;
        this.parent = p;
    ],

    method set(k, v) [
        decide [
            (k == this.key) [
                this.item = v;
            ],
            (k < this.key and this.left is System.null) [
                this.left = make System.collections._TreeNode(k, v, this);
            ],
            (k > this.key and this.right is System.null) [
                this.right = make System.collections._TreeNode(k, v, this);
            ],
            (k < this.key) [ this.left|set(k, v); ],
            (k > this.key) [ this.right|set(k, v); ]
        ];
        true;
    ],

    method find_leftmost() [
        property tmp;
        tmp = this;
        while (not (tmp.left is System.null)) [
            tmp = tmp.left;
        ];
        tmp;
    ],

    method find_rightmost() [
        property tmp;
        tmp = this;
        while (not (tmp.right is System.null)) [
            tmp = tmp.right;
        ];
        tmp;
    ],

    method remove(k) [
        property tmp;
        decide [
            (not (this.left is System.null) and this.left.key == k) [
                decide [ 
                    (not (this.left.right is System.null)) [ 
                        tmp = this.left.right|find_leftmost;
                        tmp.left = this.left.left;
                        tmp.right = this.left.right;
                        this.left = tmp;
                    ],
                    (true) [
                        this.left = this.left.left;
                    ]
                ];
            ],
            (not (this.right is System.null) and this.right.key == k) [
                decide [ 
                    (not (this.right.left is System.null)) [ 
                        tmp = this.right.left|find_rightmost;
                        tmp.left = this.right.left;
                        tmp.right = this.right.right;
                        this.right = tmp;
                    ],
                    (true) [
                        this.right = this.right.right;
                    ]
                ];
            ],
            (not (this.left is System.null) and k < this.key) [
                this.left|remove(k);
            ],
            (not (this.right is System.null) and k > this.key) [
                this.right|remove(k);
            ]
        ];
        true;
    ],

    method get(k) [
        decide [
            (k == this.key) [
                this.item;
            ],
            (k < this.key and this.left is System.null) [
                null;
            ],
            (k > this.key and this.right is System.null) [
                null;
            ],
            (k < this.key) [ this.left|get(k); ],
            (k > this.key) [ this.right|get(k); ]
        ];
    ],

    method count() [
        property x;

        x = 1;
        decide [ (not (this.left is System.null)) [ x = x + this.left|count; ] ];
        decide [ (not (this.right is System.null)) [ x = x + this.right|count; ] ];
        x;
    ]
];

##############################################################################
# Binary tree object
##############################################################################
class binary_tree [
    property root_node,
    property iterator,

    method count() [
        decide [
            (this.root_node is System.null) [ 0; ],
            (true) [ this.root_node|count; ]
        ];
    ],

    method set(key, value) [
        decide [
            (this.root_node is System.null) [ 
                this.root_node = 
                    make System.collections._TreeNode(key, value, null); 
            ],
            true [
                this.root_node|set(key, value);
            ]
        ];
        true;
    ],

    method remove(key) [
        decide [
            (this.root_node is System.null) [ true; ],
            (this.root_node.key == key) [ this.root_node = null; ],
            (true) [ this.root_node|remove(key); ]
        ];
        true;
    ],

    operator array(key) [
        decide [
            (this.root_node is System.null) [ null; ],
            (true) [ this.root_node|get(key); ]
        ];
    ],

    method reset() [
        decide [
            (this.root_node is System.null) [ null; ],
            true [ 
                this.iterator = this.root_node|find_leftmost;
                this.iterator;
            ]
        ];
    ],

    method cur() [
        decide [
            (this.iterator is System.null) [ this|reset(); ]
        ];
        this.iterator;
    ],

    method next() [
        decide [
            (this.iterator is System.null) [ false; ],
            true [
                decide [
                    (this.iterator.right is System.null) [
                        property tmp;
                        tmp = this.iterator.parent;
                        while (not (tmp is System.null) and tmp.right.key == this.iterator.key) [
                            this.iterator = tmp;
                            tmp = tmp.parent;
                        ];
                        this.iterator = tmp;
                    ],
                    true [
                        this.iterator = this.iterator.right|find_leftmost;
                    ]
                ];
                (not (this.iterator is System.null));
            ]
        ];
    ]
];

##############################################################################
# Stack object
##############################################################################
class stack [
    property list,

    construct() [
        this.list = [];
    ],

    method count() [
        this.list|count;
    ],

    method push(v) [
        this.list|prepend(v);
        true;
    ],

    method pop() [
        property tmp;
        tmp = this.list|head;
        this.list|removeAt(0);
        tmp;
    ],

    method top() [
        this.list|head;
    ]
];

##############################################################################
# Queue object
##############################################################################
class queue [
    property list,

    construct() [
        this.list = [];
    ],

    method count() [
        this.list|count;
    ],

    method enqueue(v) [
        this.list|append(v);
        true;
    ],

    method dequeue() [
        property tmp;
        tmp = this.list|head;
        this.list|removeAt(0);
        tmp;
    ]
];
