/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/
 
#include <math.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define TYPECHECK_RIGHT_OPERAND(t) \
    kite_object_t *rhs = args->builtin_data.listvalue.car; \
    kite_object_t *exc; \
    if (rhs->type != t) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Second operand is not of the requested type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define TYPECHECK_RIGHT_OPERANDS(t) \
    kite_object_t *rhs1 = args->builtin_data.listvalue.car; \
    kite_object_t *rhs2 = \
        args->builtin_data.listvalue.cdr->builtin_data.listvalue.car; \
    kite_object_t *exc; \
    if (rhs1->type != t || rhs2->type != t) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Operand(s) are not of the requested type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }
    
#define TYPECHECK_INT_OPERAND TYPECHECK_RIGHT_OPERAND(OBJ_INTEGER)
#define TYPECHECK_FLOAT_OPERAND TYPECHECK_RIGHT_OPERAND(OBJ_FLOAT)
#define TYPECHECK_FLOAT_OPERANDS TYPECHECK_RIGHT_OPERANDS(OBJ_FLOAT)

#define MATH_H_SINGLE_FLOAT_PARAM(name) \
    KITE_CLASS_METHOD(Math_ ## name) { \
        (void)this; \
        TYPECHECK_FLOAT_OPERAND; \
        kite_vm_return(thd, \
            kite_new_float(thd, name(rhs->builtin_data.floatvalue))); \
    }

#define MATH_H_SINGLE_FLOAT_PARAM_RETURN_INT(name) \
    KITE_CLASS_METHOD(Math_ ## name) { \
        (void)this; \
        TYPECHECK_FLOAT_OPERAND; \
        kite_vm_return(thd, \
            kite_new_integer(thd, name(rhs->builtin_data.floatvalue))); \
    }
    
#define MATH_H_TWO_FLOAT_PARAMS(name) \
    KITE_CLASS_METHOD(Math_ ## name) { \
        (void)this; \
        TYPECHECK_FLOAT_OPERANDS; \
        kite_vm_return(thd, \
            kite_new_float(thd, name(rhs1->builtin_data.floatvalue, \
            rhs2->builtin_data.floatvalue))); \
    }
    
#define METHOD_ADD(name, param) \
    kite_add_method(thread, newclass, # name, \
        kite_new_method(thread, FUNC_COMPILED, Math_ ## name, param));

#define CONSTANT_ADD(name, value) \
    kite_add_property(thread, newclass, # name, 1); \
    kite_set_property(thread, newclass, # name, kite_new_float(thread, value));
    
MATH_H_SINGLE_FLOAT_PARAM(acos);
MATH_H_SINGLE_FLOAT_PARAM(asin);
MATH_H_SINGLE_FLOAT_PARAM(atan);
MATH_H_TWO_FLOAT_PARAMS(atan2);
MATH_H_SINGLE_FLOAT_PARAM(ceil);
MATH_H_SINGLE_FLOAT_PARAM(cos);
MATH_H_SINGLE_FLOAT_PARAM(cosh);
MATH_H_SINGLE_FLOAT_PARAM(exp);
MATH_H_SINGLE_FLOAT_PARAM(fabs);
MATH_H_SINGLE_FLOAT_PARAM(floor);
MATH_H_TWO_FLOAT_PARAMS(fmod);
//frexp
//ldexp
MATH_H_SINGLE_FLOAT_PARAM(log);
MATH_H_SINGLE_FLOAT_PARAM(log10);
//modf
MATH_H_TWO_FLOAT_PARAMS(pow);
MATH_H_SINGLE_FLOAT_PARAM(sin);
MATH_H_SINGLE_FLOAT_PARAM(sinh);
MATH_H_SINGLE_FLOAT_PARAM(sqrt);
MATH_H_SINGLE_FLOAT_PARAM(tan);
MATH_H_SINGLE_FLOAT_PARAM(tanh);
MATH_H_SINGLE_FLOAT_PARAM(erf);
MATH_H_SINGLE_FLOAT_PARAM(erfc);
MATH_H_SINGLE_FLOAT_PARAM(gamma);
MATH_H_TWO_FLOAT_PARAMS(hypot);
MATH_H_SINGLE_FLOAT_PARAM(j0);
MATH_H_SINGLE_FLOAT_PARAM(j1);
//jn
MATH_H_SINGLE_FLOAT_PARAM(lgamma);
MATH_H_SINGLE_FLOAT_PARAM(y0);
MATH_H_SINGLE_FLOAT_PARAM(y1);
//yn
MATH_H_SINGLE_FLOAT_PARAM_RETURN_INT(isnan);
MATH_H_SINGLE_FLOAT_PARAM(acosh);
MATH_H_SINGLE_FLOAT_PARAM(asinh);
MATH_H_SINGLE_FLOAT_PARAM(atanh);
MATH_H_SINGLE_FLOAT_PARAM(cbrt);
MATH_H_SINGLE_FLOAT_PARAM_RETURN_INT(ilogb);
MATH_H_SINGLE_FLOAT_PARAM(logb);
MATH_H_TWO_FLOAT_PARAMS(nextafter);
MATH_H_TWO_FLOAT_PARAMS(remainder);
MATH_H_SINGLE_FLOAT_PARAM(rint);
MATH_H_TWO_FLOAT_PARAMS(scalb);

KITE_MODULE_INITIALIZER(math) {
    /* Create new class object. inherits is typically NULL for packages. */
    kite_object_t *newclass = kite_new_class(thread, NULL, "math");
    
    /* Add methods common to all classes. */
    METHOD_ADD(acos, 1);
    METHOD_ADD(asin, 1);
    METHOD_ADD(atan, 1);
    METHOD_ADD(atan2, 2);
    METHOD_ADD(ceil, 1);
    METHOD_ADD(cos, 1);
    METHOD_ADD(cosh, 1);
    METHOD_ADD(exp, 1);
    METHOD_ADD(fabs, 1);
    METHOD_ADD(floor, 1);
    METHOD_ADD(fmod, 2);
    //METHOD_ADD(frexp, 1);
    //METHOD_ADD(ldexp, 2);
    METHOD_ADD(log, 1);
    METHOD_ADD(log10, 1);
    //METHOD_ADD(modf, 1);
    METHOD_ADD(pow, 2);
    METHOD_ADD(sin, 1);
    METHOD_ADD(sinh, 1);
    METHOD_ADD(sqrt, 1);
    METHOD_ADD(tan, 1);
    METHOD_ADD(tanh, 1);
    METHOD_ADD(erf, 1);
    METHOD_ADD(erfc, 1);
    METHOD_ADD(gamma, 1);
    METHOD_ADD(hypot, 2);
    METHOD_ADD(j0, 1);
    METHOD_ADD(j1, 1);
    //METHOD_ADD(jn, 2);
    METHOD_ADD(lgamma, 1);
    METHOD_ADD(y0, 1);
    METHOD_ADD(y1, 1);
    //METHOD_ADD(yn, 2);
    METHOD_ADD(isnan, 1);
    METHOD_ADD(acosh, 1);
    METHOD_ADD(asinh, 1);
    METHOD_ADD(atanh, 1);
    METHOD_ADD(cbrt, 1);
    METHOD_ADD(ilogb, 1);
    METHOD_ADD(logb, 1);
    METHOD_ADD(nextafter, 2);
    METHOD_ADD(remainder, 2);
    METHOD_ADD(rint, 1);
    METHOD_ADD(scalb, 2);
    
    /* Add constants */
    CONSTANT_ADD(E, M_E);
    CONSTANT_ADD(LOG2E, M_LOG2E);
    CONSTANT_ADD(LOG10E, M_LOG10E);
    CONSTANT_ADD(LN2, M_LN2);
    CONSTANT_ADD(LN10, M_LN10);
    CONSTANT_ADD(PI, M_PI);
    CONSTANT_ADD(PI_2, M_PI_2);
    CONSTANT_ADD(PI_4, M_PI_4);
    CONSTANT_ADD(ONE_PI, M_1_PI);
    CONSTANT_ADD(TWO_PI, M_2_PI);
    CONSTANT_ADD(TWO_SQRTPI, M_2_SQRTPI);
    CONSTANT_ADD(SQRT2, M_SQRT2);
    CONSTANT_ADD(SQRT1_2, M_SQRT1_2);
    
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "math", 1);
    kite_set_property(thread, parent, "math", newclass);
}

